// pam_url - GPLv2, Sascha Thomas Spreitzer, https://fedorahosted.org/pam_url

#include "pam_url.h"

#include <stdio.h>
#include <stdint.h>

char* recvbuf = NULL;
size_t recvbuf_size = 0;

void debug(pam_handle_t* pamh, const char *msg)
{
	pam_syslog(pamh, LOG_ERR, "%s", msg);
}

int get_password(pam_handle_t* pamh, pam_url_opts* opts)
{
	char* p = NULL;
	char* prompt;
	int prompt_len = 0;

	config_t config;
	config_init(&config);
	config_read_file(&config,"/etc/pam_url.conf");

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.settings.prompt", (const char **) &prompt) )
	{
		prompt_len = strlen(DEF_PROMPT) + 1;
		prompt = calloc(1, prompt_len);
		if (prompt == NULL)
			return PAM_AUTH_ERR;
		strncpy(prompt, DEF_PROMPT, prompt_len);
    }

	pam_prompt(pamh, PAM_PROMPT_ECHO_OFF, &p, "%s", prompt);
	free(prompt);
	if( NULL != p && strlen(p) > 0)
	{
		opts->passwd = p;
		return PAM_SUCCESS;
	}
	else
	{
		return PAM_AUTH_ERR;
	}
}

int parse_opts(pam_url_opts* opts, int argc, const char** argv, int mode)
{
	int i = 0;
	int len = 0;

#ifdef DEBUG
	pam_url_debug = 1;
#else
	pam_url_debug = 0;
#endif

	if( 0 != argc && NULL != argv)
	{
		for( i = 0; i < argc; i++)
		{
			if( 0 == strcmp(argv[i], "debug") )
			{
				pam_url_debug = 1;
			}

			else if( 0 == strncmp(argv[i], "config=", strlen("config=")) )
			{
				len = strlen(argv[i]) - strlen("config=") + 1;
				opts->configfile = calloc(1, len);
				if (opts->configfile == NULL)
					goto pam_error;
				strncpy(opts->configfile, argv[i] + strlen("config="), len);
			}
		}
	}

	switch(mode)
	{
		case PAM_SM_ACCOUNT:
			len = strlen("PAM_SM_ACCOUNT") + 1;
			opts->mode = calloc(1, len);
				if (opts->mode == NULL)
					goto pam_error;
			strncpy(opts->mode, "PAM_SM_ACCOUNT", len);
			break;

		case PAM_SM_SESSION:
			len = strlen("PAM_SM_SESSION") + 1;
			opts->mode = calloc(1, len);
				if (opts->mode == NULL)
					goto pam_error;
			strncpy(opts->mode, "PAM_SM_SESSION", len);
			break;

		case PAM_SM_PASSWORD:
			len = strlen("PAM_SM_PASSWORD") + 1;
			opts->mode = calloc(1, len);
				if (opts->mode == NULL)
					goto pam_error;
			strncpy(opts->mode, "PAM_SM_PASSWORD", len);
			break;

		default: // PAM_SM_AUTH
			len = strlen("PAM_SM_AUTH") + 1;
			opts->mode = calloc(1, len);
				if (opts->mode == NULL)
					goto pam_error;
			strncpy(opts->mode,"PAM_SM_AUTH", len);
	}

	config_t config;
	config_init(&config);
	config_read_file(&config,"/etc/pam_url.conf");

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.settings.uri", (const char **) &opts->url) )
	{
		len = strlen(DEF_URL) + 1;
		opts->url = calloc(1, len);
		if (opts->url == NULL)
			goto pam_error;
		strncpy(opts->url, DEF_URL, len);
	}

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.settings.returncode", (const char **) &opts->RC) )
	{
		len = strlen(DEF_RETURNCODE) + 1;
		opts->RC = calloc(1, len);
		if (opts->RC == NULL)
			goto pam_error;
		strncpy(opts->RC, DEF_RETURNCODE, len);
	}

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.settings.userfield", (const char **) &opts->userfield) )
	{
		len = strlen(DEF_USER) + 1;
		opts->userfield = calloc(1, len);
		if (opts->userfield == NULL)
			goto pam_error;
		strncpy(opts->userfield, DEF_USER, len);
	}

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.settings.passwdfield", (const char **) &opts->passwdfield) )
	{
		len = strlen(DEF_PASSWD) + 1;
		opts->passwdfield = calloc(1, len);
		if (opts->passwdfield == NULL)
			goto pam_error;
		strncpy(opts->passwdfield, DEF_PASSWD, len);
	}

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.settings.extradata", (const char **) &opts->extrafield) )
	{
		len = strlen(DEF_EXTRA) + 1;
		opts->extrafield = calloc(1, len);
		if (opts->extrafield == NULL)
			goto pam_error;
		strncpy(opts->extrafield, DEF_EXTRA, len);
	}

	if( CONFIG_FALSE == config_lookup_bool(&config, "pam_url.ssl.verify_peer", &opts->ssl_verify_peer) )
	{
		opts->ssl_verify_peer = 1;
	}

	if( CONFIG_FALSE == config_lookup_bool(&config, "pam_url.ssl.verify_host", &opts->ssl_verify_host) )
	{
		opts->ssl_verify_host = 1;
	}

	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.ssl.client_cert", (const char **) &opts->sslcert) )
	{
		len = strlen(DEF_SSLCERT) + 1;
		opts->sslcert = calloc(1, len);
		if (opts->sslcert == NULL)
			goto pam_error;
		strncpy(opts->sslcert, DEF_SSLCERT, len);
	}


	if( CONFIG_FALSE == config_lookup_string(&config, "pam_url.ssl.client_key", (const char **) &opts->sslkey) )
	{
		len = strlen(DEF_SSLKEY) + 1;
		opts->sslkey = calloc(1, len);
		if (opts->sslkey == NULL)
			goto pam_error;
		strncpy(opts->sslkey, DEF_SSLKEY, len);
	}


	return PAM_SUCCESS;

pam_error:
    return PAM_SYSTEM_ERR;
}

size_t curl_wf(void *ptr, size_t size, size_t nmemb, void *stream)
{
	size_t oldsize=0;

	if( 0 == size * nmemb )
		return 0;

	if( NULL == recvbuf )
	{
		if( NULL == ( recvbuf = calloc(nmemb, size) ) )
		{
			return 0;
		}
	}

	// Check the multiplication for an overflow
	if (((nmemb * size) > (SIZE_MAX / nmemb)) ||
			// Check the addition for an overflow
			((SIZE_MAX - recvbuf_size) < (nmemb * size))) {
		// The arithmetic will cause an integer overflow
		return 0;
	}
	if( NULL == ( recvbuf = realloc(recvbuf, recvbuf_size + (nmemb * size)) ) )
	{
		return 0;
	}
	else
	{
		oldsize = recvbuf_size;
		recvbuf_size += nmemb * size;
		memcpy(recvbuf + oldsize, ptr, size * nmemb);
		return(size*nmemb);
	}
}

int curl_debug(CURL *C, curl_infotype info, char * text, size_t textsize, void* pamh)
{
	debug((pam_handle_t*)pamh, text);
	return 0;
}

int fetch_url(pam_handle_t *pamh, pam_url_opts opts)
{
	CURL* eh = NULL;
	char* post = NULL;
	int ret = 0;

	if( NULL == opts.user )
		opts.user = calloc(1,1);
		if (opts.user == NULL)
			goto curl_error;

	if( NULL == opts.passwd )
		opts.passwd = calloc(1,1);
		if (opts.passwd == NULL)
			goto curl_error;

	ret = asprintf(&post, "%s=%s&%s=%s&mode=%s%s", opts.userfield,
													(char*)opts.user,
													opts.passwdfield,
													(char*)opts.passwd,
													opts.mode,
													opts.extrafield);

	if (ret == -1)
		// If this happens, the contents of post are undefined, we could
		// end up freeing an uninitialized pointer, which could crash (but
		// should not have security implications in this context).
		goto curl_error;

	if( 0 != curl_global_init(CURL_GLOBAL_ALL) )
		goto curl_error;

	if( NULL == (eh = curl_easy_init() ) )
		goto curl_error;

	if( 1 == pam_url_debug)
	{
		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_VERBOSE, 1) )
			goto curl_error;

		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_DEBUGDATA, pamh) )
			goto curl_error;

		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_DEBUGFUNCTION, curl_debug) )
			goto curl_error;
	}

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_POSTFIELDS, post) )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_USERAGENT, USER_AGENT) )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_WRITEFUNCTION, curl_wf) )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_URL, opts.url) )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSLCERT, opts.sslcert) )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSLCERTTYPE, "PEM") )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSLKEY, opts.sslkey) )
		goto curl_error;

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSLKEYTYPE, "PEM") )
		goto curl_error;

	if( opts.ssl_verify_host == 1 )
	{
		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSL_VERIFYHOST, 2) )
			goto curl_error;
	}
	else
	{
		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSL_VERIFYHOST, 0) )
			goto curl_error;
	}

	if( opts.ssl_verify_peer == 1 )
	{
		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSL_VERIFYPEER, 1) )
			goto curl_error;
	}
	else
	{
		if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_SSL_VERIFYPEER, 0) )
			goto curl_error;
	}

	if( CURLE_OK != curl_easy_setopt(eh, CURLOPT_FAILONERROR, 1) )
		goto curl_error;

	if( CURLE_OK != curl_easy_perform(eh) )
		goto curl_error;

	// No errors
	curl_easy_cleanup(eh);
	free(post);
	return PAM_SUCCESS;

curl_error:
	if (eh != NULL)
		curl_easy_cleanup(eh);
	if (post != NULL)
		free(post);
	return PAM_AUTH_ERR;
}

int check_rc(pam_url_opts opts)
{
	int ret=0;

	if( NULL == recvbuf )
	{
		ret++;
		return PAM_AUTH_ERR;
	}

	if( 0 != memcmp(opts.RC, recvbuf, strlen(opts.RC)) )
		ret++;

	if( 0 != ret )
	{
		return PAM_AUTH_ERR;
	}
	else
	{
		return PAM_SUCCESS;
	}
}

void cleanup(pam_url_opts* opts)
{
	if( NULL != recvbuf )
		free(recvbuf);

	recvbuf_size=0;
}
